//
// SPDX-License-Identifier: BSD-3-Clause
// Copyright (c) Contributors to the OpenEXR Project.
//

#ifndef NAMESPACEALIAS_H_
#define NAMESPACEALIAS_H_

#include <ImfNamespace.h>

namespace IMF = OPENEXR_IMF_NAMESPACE;

#endif /* NAMESPACEALIAS_H_ */
